@@ -3,6 +3,7 @@
 from __future__ import with_statement
 from binascii import hexlify
 import atexit
+import math
 import os
 import pipes
 import re
@@ -23,10 +24,10 @@ from debughelpers import (Error, Info, UnloggedError, UnloggedInfo,
 						  UnloggedWarning, Warn)
 from log import LogFile, safe_print
 from meta import name as appname
-from multiprocess import mp
+from multiprocess import mp, pool_slice
 from options import debug, verbose
 from util_os import getenvu, quote_args, which
-from util_str import safe_basestring, safe_str, safe_unicode
+from util_str import make_filename_safe, safe_basestring, safe_str, safe_unicode
 import CGATS
 import colormath
 import config
@@ -38,6 +39,38 @@ def Property(func):
 	return property(**func())
 
 
+def _mp_xicclu(chunk, thread_abort_event, progress_queue, profile_filename,
+			   intent="r", direction="f", order="n",
+			   pcs=None, scale=1, cwd=None, startupinfo=None, use_icclu=False,
+			   use_cam_clipping=False, logfile=None,
+			   show_actual_if_clipped=False, input_encoding=None,
+			   output_encoding=None, abortmessage="Aborted"):
+	if not config.cfg.items(config.ConfigParser.DEFAULTSECT):
+		config.initcfg()
+	profile = ICCP.ICCProfile(profile_filename)
+	xicclu = Xicclu(profile, intent, direction, order, pcs, scale, cwd,
+					startupinfo, use_icclu, use_cam_clipping, logfile,
+					None, show_actual_if_clipped, input_encoding,
+					output_encoding)
+	prevperc = 0
+	start = 0
+	num_subchunks = 50
+	subchunksize = float(len(chunk)) / num_subchunks
+	for i in xrange(num_subchunks):
+		if thread_abort_event.is_set():
+			xicclu.exit()
+			return Info(abortmessage)
+		end = int(math.ceil(subchunksize * (i + 1)))
+		xicclu(chunk[start:end])
+		start = end
+		perc = round((i + 1.0) / num_subchunks * 100)
+		if progress_queue and perc > prevperc:
+			progress_queue.put(perc - prevperc)
+			prevperc = perc
+	xicclu.exit()
+	return xicclu.get()
+
+
 def _mp_generate_B2A_clut(chunk, thread_abort_event, progress_queue,
 						  profile_filename, intent, direction, pcs,
 						  use_cam_clipping, clutres, step, threshold,
@@ -71,15 +104,19 @@ def _mp_generate_B2A_clut(chunk, thread_abort_event, progress_queue,
 	for interp_tuple in (interp, Linterp):
 		if interp_tuple:
 			# Use numpy for speed
-			interp_list = list(interp_tuple)
+			if interp_tuple is interp:
+				interp_list = list(interp_tuple)
+			else:
+				interp_list = [interp_tuple]
 			for i, ointerp in enumerate(interp_list):
 				interp_list[i] = colormath.Interp(ointerp.xp, ointerp.fp,
 												  use_numpy=True)
 			if interp_tuple is interp:
 				interp = interp_list
 			else:
-				Linterp = interp_list
-	m2i = m2.inverted()
+				Linterp = interp_list[0]
+	if profile.connectionColorSpace == "XYZ":
+		m2i = m2.inverted()
 	for a in chunk:
 		if thread_abort_event.is_set():
 			if use_cam_clipping:
@@ -425,12 +462,23 @@ class Xicclu(WorkerBase):
 			else:
 				profile = ICCP.ICCProfile(profile)
 		is_profile = isinstance(profile, ICCP.ICCProfile)
+		if is_profile:
+			if profile.fileName:
+				prefix = os.path.basename(profile.fileName)
+			else:
+				prefix = make_filename_safe(profile.getDescription(),
+											concat=False) + profile_ext
+			prefix += "-"
+			if (profile.version >= 4 and
+				not profile.convert_iccv4_tags_to_iccv2()):
+				raise Error("\n".join([lang.getstr("profile.iccv4.unsupported"),
+												   profile.getDescription()]))
 		if not profile.fileName or not os.path.isfile(profile.fileName):
 			if not cwd:
 				cwd = self.create_tempdir()
 				if isinstance(cwd, Exception):
 					raise cwd
-			fd, profile.fileName = tempfile.mkstemp(profile_ext, dir=cwd)
+			fd, profile.fileName = tempfile.mkstemp("", prefix, dir=cwd)
 			stream = os.fdopen(fd, "wb")
 			profile.write(stream)
 			stream.close()
@@ -449,11 +497,13 @@ class Xicclu(WorkerBase):
 		xicclu = safe_str(xicclu)
 		cwd = safe_str(cwd)
 		args = [xicclu, "-s%s" % scale]
+		self.show_actual_if_clipped = False
 		if utilname == "xicclu":
 			if (is_profile and
 				show_actual_if_clipped and "A2B0" in profile.tags and
 				("B2A0" in profile.tags or direction == "if")):
 				args.append("-a")
+				self.show_actual_if_clipped = True
 			if use_cam_clipping:
 				args.append("-b")
 			if get_argyll_version("xicclu") >= [1, 6]:
@@ -466,13 +516,23 @@ class Xicclu(WorkerBase):
 					output_encoding = "n"
 				args += ["-e" + input_encoding, "-E" + output_encoding]
 		args.append("-f" + direction)
+		self.output_scale = 1.0
 		if is_profile:
 			if profile.profileClass not in ("abst", "link"):
 				args.append("-i" + intent)
 				if order != "n":
 					args.append("-o" + order)
-			if pcs and profile.profileClass != "link":
-				args.append("-p" + pcs)
+			if profile.profileClass != "link":
+				if (direction in ("f", "ib") and
+					(pcs == "x" or (profile.connectionColorSpace == "XYZ" and
+									not pcs))):
+					# In case of forward lookup with XYZ PCS, use 0..100 scaling
+					# internally so we get extra precision from xicclu for the
+					# decimal part. Scale back to 0..1 later.
+					pcs = "X"
+					self.output_scale = 100.0
+				if pcs:
+					args.append("-p" + pcs)
 		args.append(self.profile_path)
 		if debug or verbose > 1:
 			self.sessionlogfile = LogFile(profile_basename + ".xicclu",
@@ -494,9 +554,7 @@ class Xicclu(WorkerBase):
 					indent += " "
 				self.log("")
 			self.log(lang.getstr("commandline"))
-			printcmdline(xicclu if debug or verbose > 2 else
-						 os.path.basename(xicclu), args[1:], fn=self.log,
-						 cwd=cwd)
+			printcmdline(xicclu, args[1:], fn=self.log, cwd=cwd)
 			self.log("")
 		self.startupinfo = startupinfo
 		self.args = args
@@ -617,7 +675,12 @@ class Xicclu(WorkerBase):
 		for i, line in enumerate(self.output):
 			line = line.strip()
 			if line.startswith("["):
-				if self.sessionlogfile:
+				if parsed and get_clip and self.show_actual_if_clipped:
+					parts = line.strip("[]").split(",")
+					actual = [float(v) for v in parts[0].split()[1:4]]  # Actual CIE
+					actual.append(float(parts[1].split()[-1]))  # deltaE
+					parsed[-1].append(actual)
+				elif self.sessionlogfile:
 					self.sessionlogfile.write(line)
 				continue
 			elif not "->" in line:
@@ -630,8 +693,8 @@ class Xicclu(WorkerBase):
 			clip = parts.pop() == "(clip)"
 			if clip:
 				parts.pop()
-			parsed.append([float(n) for n in parts])
-			if get_clip:
+			parsed.append([float(n) / self.output_scale for n in parts])
+			if get_clip and not self.show_actual_if_clipped:
 				parsed[-1].append(clip)
 			j += 1
 		if self.sessionlogfile:
@@ -661,3 +724,55 @@ class Xicclu(WorkerBase):
 			pass
 		
 		return locals()
+
+
+class MP_Xicclu(Xicclu):
+
+	def __init__(self, profile, intent="r", direction="f", order="n",
+				 pcs=None, scale=1, cwd=None, startupinfo=None, use_icclu=False,
+				 use_cam_clipping=False, logfile=None, worker=None,
+				 show_actual_if_clipped=False, input_encoding=None,
+				 output_encoding=None):
+		WorkerBase.__init__(self)
+		self.logfile = logfile
+		self.worker = worker
+		self._in = []
+		self._args = (profile.fileName, intent, direction, order,
+					  pcs, scale, cwd, startupinfo, use_icclu,
+					  use_cam_clipping, None,
+					  show_actual_if_clipped, input_encoding,
+					  output_encoding, lang.getstr("aborted"))
+		self._out = []
+		num_cpus = mp.cpu_count()
+		if isinstance(profile.tags.get("A2B0"), ICCP.LUT16Type):
+			size = profile.tags.A2B0.clut_grid_steps
+			self.num_workers = min(max(num_cpus, 1), size)
+			if num_cpus > 2:
+				self.num_workers = int(self.num_workers * 0.75)
+			self.num_batches = size // 9
+		else:
+			if num_cpus > 2:
+				self.num_workers = 2
+			else:
+				self.num_workers = num_cpus
+			self.num_batches = 1
+
+	def __call__(self, idata):
+		self._in.append(idata)
+
+	def close(self):
+		pass
+
+	def exit(self):
+		pass
+
+	def spawn(self):
+		pass
+
+	def get(self):
+
+		for slices in pool_slice(_mp_xicclu, self._in, self._args, {},
+								 self.num_workers, self.thread_abort,
+								 self.logfile, num_batches=self.num_batches):
+			self._out.extend(slices)
+		return self._out
